#!/usr/bin/env python
# Copyright (c) 2009, 2010, 2011 Arista Networks, Inc.  All rights reserved.
# Arista Networks, Inc. Confidential and Proprietary.

import Ethernet
import struct
import socket
import Tracing
from if_ether_arista import TC_PRIO_CONTROL 
from zlib import crc32

hbt0 = Tracing.trace0 # Heartbeat header check

MLAG_PORT = 4432
# Used during reload-delay to transfer ARP table from Primary->Secondary
MLAG_ARP_SYNC_PORT = 50002
# Set Mlag socket TTL to 255
# to pass through Acl filter for Mlag
MLAG_TTL = 255
# Set number of SYN retries to be low so tacc retries the
# connection more often than the kernel backoff algorithm
MLAG_SYN_RETRIES = 1
# Set Mlag socket TOS to 0xe0
MLAG_TOS = 0xe0

CALLBACK_HIGH_PRIORITY = 0
CALLBACK_MEDIUM_PRIORITY = 1
CALLBACK_LOW_PRIORITY = 2

HEARTBEAT_TIMEOUT_MULTIPLIER = 15
HEARTBEAT_INITIAL_WAIT = 300.0
MAX_HEARTBEAT_TIMEOUT = 75000

UDPHEARTBEATPORTID = 4096
UDPHEARTBEAT_HDR_MAGICNUM_NO_TLV = 547 # <=V7
UDPHEARTBEAT_HDR_MAGICNUM = 548        # >= V8
UDPHEARTBEAT_HDR_FORMAT = "!HIH"
UDPHEARTBEAT_HDR_SIZE = struct.calcsize( UDPHEARTBEAT_HDR_FORMAT )

VERSION_PKT_FORMAT = "!II"
VERSION_PKT_SIZE = struct.calcsize( VERSION_PKT_FORMAT )

PEER_TIME_OFFSET_FORMAT = "!ddd"
PEER_TIME_OFFSET_SIZE = struct.calcsize( PEER_TIME_OFFSET_FORMAT )

NEG_STATUS_ALL = ( "Connected",
                   "Connecting",
                   "Domain mismatch",
                   "Invalid peer",
                   "In negotiation",
                   "Version incompatible",
                   "Hardware not ready" )
( NEG_STATUS_CONNECTED,
  NEG_STATUS_CONNECTING,
  NEG_STATUS_DOMAIN_MISMATCH,
  NEG_STATUS_INVALID_PEER,
  NEG_STATUS_NEGOTIATION,
  NEG_STATUS_VERS_INCOMPATIBLE,
  NEG_STATUS_HW_NOT_READY ) = NEG_STATUS_ALL

def configured( cfg ):
   return bool( cfg.enabled == True and \
                cfg.domainId != '' and \
                cfg.localInterface and
                not cfg.localInterface.isOrphan() and \
                cfg.peerAddress != '0.0.0.0' and \
                cfg.peerLinkIntf and
                not cfg.peerLinkIntf.isOrphan() )

def heartbeatTimeoutFromInterval( interval ):
   return min( HEARTBEAT_TIMEOUT_MULTIPLIER * interval,
               MAX_HEARTBEAT_TIMEOUT )

def heartbeatTimeout( cfg, interval ):
   return cfg.heartbeatTimeout \
       if cfg.heartbeatTimeout != 0 \
       else heartbeatTimeoutFromInterval( interval )

def msi( macAddr ):
   # Set locally administered bit
   mc = Ethernet.convertMacAddrToPackedString( macAddr )
   # We don't expect the bit to be set already.
   assert ord( mc[0] ) & 0x2 == 0
   mc = chr( ord( mc[0] ) | 0x2 ) + mc[ 1: ]
   colonAddr = '%02x:%02x:%02x:%02x:%02x:%02x' % tuple( [ ord( b ) for b in mc ] )
   return colonAddr

def minSystemId( macAddrs ):
   return min( [ msi( macAddr ) for macAddr in macAddrs ] )

def newVersionPkt( minVersion, maxVersion ):
   versionPkt = struct.pack( VERSION_PKT_FORMAT, minVersion, maxVersion )
   return versionPkt

def versionFieldsFromPkt( pkt ):
   assert len( pkt ) >= VERSION_PKT_SIZE
   versionPkt = pkt[ : VERSION_PKT_SIZE ]
   minVersion, maxVersion = struct.unpack( VERSION_PKT_FORMAT,
                                           versionPkt )
   return ( minVersion, maxVersion )

def newHeartbeatHdr( domainId, useTlv=True ):
   domainIdCrc = crc32( domainId ) & 0xffffffff
   heartbeatHdr = struct.pack( UDPHEARTBEAT_HDR_FORMAT, 
                               UDPHEARTBEAT_HDR_MAGICNUM if useTlv else
                               UDPHEARTBEAT_HDR_MAGICNUM_NO_TLV,
                               domainIdCrc,
                               # The lsb for this field denotes whether the
                               # packet came via peer's local interface (which
                               # is always false for UDP heartbeat packet)
                               # hence the left shift.
                               UDPHEARTBEATPORTID << 1 )
   return heartbeatHdr

def validateHeartbeatPkt( pktStr, domainId ):
   if len( pktStr ) < UDPHEARTBEAT_HDR_SIZE:
      hbt0( "Received packet with invalid length:", len( pktStr ) )
      return
   hdr = pktStr[ : UDPHEARTBEAT_HDR_SIZE ]
   ( magicNum, domainIdCrc, fromPeerIntfAndPortId ) = \
         struct.unpack( UDPHEARTBEAT_HDR_FORMAT, hdr )
   if ( magicNum != UDPHEARTBEAT_HDR_MAGICNUM_NO_TLV and
        magicNum != UDPHEARTBEAT_HDR_MAGICNUM ):
      hbt0( "Received packet with unexpected magic number: ", magicNum )
      return
   localDomainIdCrc = crc32( domainId ) & 0xffffffff
   if domainIdCrc != localDomainIdCrc:
      hbt0( "Received packet with unexpected domainId CRC: ", domainIdCrc, 
            " local: ", localDomainIdCrc )
      return
   portId = fromPeerIntfAndPortId >> 1
   if portId != UDPHEARTBEATPORTID:
      hbt0( "Received packet with unexpected portId: ", portId )
      return
   return pktStr[ UDPHEARTBEAT_HDR_SIZE : ]

def socketOpts( synCntOption=True, deviceName=None ):
   # security and QoS options for Mlag sockets
   opts = [ ( socket.IPPROTO_IP, socket.IP_TTL, MLAG_TTL ),
            ( socket.IPPROTO_IP, socket.IP_TOS, MLAG_TOS ),
            ( socket.SOL_SOCKET, socket.SO_PRIORITY, TC_PRIO_CONTROL ) ]
   if synCntOption:
      opts += [ ( socket.IPPROTO_TCP, socket.TCP_SYNCNT, MLAG_SYN_RETRIES ) ]
   if deviceName:
      opts += localInterfaceSocketOpts( deviceName )
   return opts

def localInterfaceSocketOpts( deviceName ):
   # Socket options specific to the local-interface
   return [ ( socket.SOL_SOCKET, socket.SO_BINDTODEVICE, deviceName ) ]


